/*
**   _______              __
**  / ___/ /  ___  __ _  / /  ___
** / /__/ _ \/ _ \/  ' \/ _ \/ _ \
** \___/_//_/\___/_/_/_/_.__/\___/ 
**
** This software is copyright (C) by the Lawrence Berkeley
** National Laboratory.  Permission is granted to reproduce
** this software for non-commercial purposes provided that
** this notice is left intact.
** 
** It is acknowledged that the U.S. Government has rights to
** this software under Contract DE-AC03-765F00098 between
** the U.S. Department of Energy and the University of
** California.
**
** This software is provided as a professional and academic
** contribution for joint exchange.  Thus it is experimental,
** is provided ``as is'', with no warranties of any kind
** whatsoever, no support, no promise of updates, or printed
** documentation.  By using this software, you acknowledge
** that the Lawrence Berkeley National Laboratory and
** Regents of the University of California shall have no
** liability with respect to the infringement of other
** copyrights by any part of this software.
**
*/
// Author: Ted Sternberg

#include "ChomboHDF5.h"


/** This is called very early on, as it finds items which determine what
 *  ChomboHDF5 template parameters and subclass we need to instantiate.
*/
Status
ChomboHDF5DiscoverMetaparameters(
    char const * infileName,
    int * dimensionality,          // 2 or 3
    int * precision,               // 1 (float) or 2 (double)
    ChomboHDF5FileType * filetype )
{
    Trace t("ChomboHDF5DiscoverMetaparameters()");
    Status status( STATUS_OK );

    //
    // Open file.
    //
    hid_t fileHID = H5Fopen( infileName, H5F_ACC_RDONLY, H5P_DEFAULT);
    if( fileHID < 0 )
    {
        status = STATUS_BAD_FILE;
        t.Error( "Failed to open file %s.", infileName );
        return status;
    }

    {
      HDF5GroupID chombo_global_GroupHID( fileHID, Consts::Chombo_global );
  
      //
      // Dimensionality
      //
      hid_t attribute = H5Aopen_name( chombo_global_GroupHID.GetHid(),
                                      Consts::SpaceDim );
      herr_t ret  = H5Aread( attribute, H5T_NATIVE_INT, dimensionality );
      if( ret != 0 )
      {
          status = STATUS_BAD_FILE;
          t.Error( "Failed to find attribute %s.", Consts::SpaceDim );
          return status;
      }
      if( (*dimensionality != 2) && (*dimensionality != 3) )
      {
          status = STATUS_BAD_FILE;
          t.Error( "Illegal dimensionality: %d", *dimensionality );
          return status;
      }
      H5Aclose( attribute );
  
      //
      // Precision
      //
      attribute =
          H5Aopen_name( chombo_global_GroupHID.GetHid(), Consts::testReal );
      hid_t datatype = H5Aget_type( attribute );
  
      if( H5Tget_precision( datatype ) == H5Tget_precision( H5T_NATIVE_FLOAT ) )
      {
          *precision = 1;
      } else if( H5Tget_precision( datatype ) == 
                 H5Tget_precision( H5T_NATIVE_DOUBLE ) )
      {
          *precision = 2;
      } else
      {
          status = STATUS_BAD_FILE;
          t.Error( "Illegal datatype -- neither float nor double." );
          return status;
      }

      H5Aclose( attribute );
      H5Tclose( datatype );
    }

    //
    // Filetype -- old Chombo or new EBChombo.  No point looking at the
    // "filetype" attribute, as it's spelled with 'f' in the old format, and
    // with 'F' in the new format.  So we just go by whether there's a small or
    // capital 'f' there!
    //
    {
      HDF5GroupID root_GroupHID( fileHID, "/" );
      EBConsts  ebConsts;
      if( root_GroupHID.ContainsItem( string(ebConsts.filetype) ) )
      {
          *filetype = new_eb_chombo;
      } else
      {
          *filetype = old_chombo;
      }
    }

    H5Fclose( fileHID );
    return status;
}


//
// Convenience function (and two specializations) for use in AttributeScan.
//
// Interesting: if we put the general template in the .hImpl file, then the
// template specializations (in the .cpp file) don't get instantiated and the
// flow of control goes to the assert(0) here.  But with the general template
// here, none of them get instantiated
template<class REAL_T> herr_t ChomboH5Aread( hid_t attr, REAL_T * x )
{
    assert(0); // <REAL_T> other than float or double is illegal.
}

template<> herr_t ChomboH5Aread<float>( hid_t attr, float * x )
{
    return H5Aread( attr, H5T_NATIVE_FLOAT, x );
}
template<> herr_t ChomboH5Aread<double>( hid_t attr, double * x )
{
    return H5Aread( attr, H5T_NATIVE_DOUBLE, x );
}


string ReadHDF5StringAttribute( HDF5GroupID const & group,
                                string attributeName,
                                Status * status )
{
    hid_t attrHID = H5Aopen_name( group.GetHid(), attributeName.c_str() );
    *status = STATUS_OK;
    if( attrHID <= 0 )
    {
        *status = STATUS_NOT_FOUND;
        return ""; // HDF5 has just printed lots of warning verbiage for us.
    }
    hid_t atype  = H5Aget_type(attrHID);
    size_t size = H5Tget_size(atype);
    char * buf = new char[size+1];
    H5Aread(attrHID, atype, buf);
    buf[size]=0;    

    H5Aclose( attrHID );
    H5Tclose( atype );
    assert( buf );
    return string(buf);
}



//----------- Methods of HDF5GroupID ------------------
//
/** Opens the group.  Prints error message if something goes wrong.
 *  If you try to open a nonexistent group, hdf5 will complain verbosely.  The
 *  right way to check if a group exists is with H5Gget_objinfo().
 */
HDF5GroupID::HDF5GroupID( hid_t parentHid, char const * groupName )
  : m_parentHid( parentHid ),
    m_groupName( groupName )
{
    Trace t("HDF5GroupID::HDF5GroupID()");
    m_hid = H5Gopen( parentHid, groupName );
    if( m_hid < 0 )
    {
        t.Error( "Failed to open HDF5 group %s.", groupName );
        // But not to worry, HDF5 prints a whole lot of complaints of its own
        // if we get here.
    }
}

/** Closes the group */
HDF5GroupID::~HDF5GroupID()
{
    H5Gclose( m_hid );
}


bool
HDF5GroupID::ContainsItem( string itemName ) const
{
    Trace t("HDF5GroupID::ContainsItem()");
    
    string itemNameCopy( itemName );

    H5Giterate( m_parentHid, m_groupName.c_str(), 0, FindGroupItem,
                &itemNameCopy );
    if( itemNameCopy == string(Consts::unlikely_name) )
    {   
        return true;
    } else
    {   // H5Giterate discovers sub-groups and datasets but not attributes, so
        // we have to turn now to H5Aiterate:
        H5Aiterate( m_hid, 0, FindGroupItem,
                    &itemNameCopy );
        if( itemNameCopy == string(Consts::unlikely_name) )
        {
            return true;
        } else
        {
            return false;
        }
    }


    /* Here's a tempting way to do it, unfortunately it discovers sub-groups and
     * dataseta but not attributes:
    DisableHDF5ErrorPrinting();
    herr_t ret = H5Gget_objinfo( m_hid, itemName.c_str(), 0,0 );
    ReEnableHDF5ErrorPrinting();
    return (! ret);
    */    
}

/** Passed to H5Giterate, to help HDF5GroupID::ContainsItem() determine if a
 *  group contains an item of a certain name.
 *  Arg soughtItem is really a string* (and it comes through the last argument
 *  passed to H5Giterate).  If it matches currItem, then we set
 *  *soughtItem to Consts::unlikely_name -- that's how we indicate soughtItem
 *  has been found.  FIXME: There must be a more elegant way to do this; what
 *  role does the return value play??
*/
herr_t
FindGroupItem( hid_t loc_id, char const * currItem, void * soughtItem  )
{
    //cerr << "@@ Found group item:|" << currItem << "|" << endl;
    string * soughtItemAsString( static_cast<string *>(soughtItem) );

    if( string(currItem) == *soughtItemAsString )
    {
        *soughtItemAsString = string(Consts::unlikely_name);
    }

    return 0;
}


/** Returns a vector of the names of all the items under this group. */
vector<string>
HDF5GroupID::FindAllItems() const
{
    Trace t("HDF5GroupID::FindAllItems()");
    vector<string> result;
    H5Giterate( m_parentHid, m_groupName.c_str(), 0, AccumulateGroupItems,
                &result );
    return result;
}

/** FindAllItems() passes this to H5Giterate().
 *  Arg stringVect is actually a pointer to a vector<string>.  We use that to
 *  accumulate the currItem every time we enter this function.
 *  H5Giterate seems to find DATASET and GROUP items, but not ATTRIBUTE items.
*/
herr_t
AccumulateGroupItems( hid_t loc_id, char const * currItem, void * stringVect  )
{
    vector<string> * stringVectAsVector(
        static_cast<vector<string> *>(stringVect) );
    stringVectAsVector->push_back( currItem );
    return 0;
}



/** The HDF5 file uses an int to indicate data centering, while everywhere else
 *  in ChomboVis we use an Intvect.  Here's the conversion table for 3D.
 *   0 => 0,0,0 : cell-centered
 *   1 => 1,0,0 : x-face-centered
 *   2 => 0,1,0 : y-face-centered
 *   3 => 0,0,1 : z-face-centered
 *   4 => 0,1,1 : x-edge-centered
 *   5 => 1,0,1 : y-edge-centered
 *   6 => 1,1,0 : z-edge-centered
 *   7 => 1,1,1 : node-centered
 *
 * In 2D, the only legal choices are 0,1,2 and 7.
*/
Intvect DataCenteringInt2Intvect( int dc, int spaceDim )
{
    Trace t("DataCenteringInt2Intvect()");
    if( spaceDim == 2 )
    {
        switch( dc )
        {
            case 0 : return Intvect(0,0,0);
            case 1 : return Intvect(1,0,0);
            case 2 : return Intvect(0,1,0);
            case 7 : return Intvect(1,1,0);
            default: t.FatalError( "Illegal data centering code: %d.", dc );
        }
    }
    else /* spaceDim == 3 */
    {
        switch( dc )
        {
            case 0 : return Intvect(0,0,0);
            case 1 : return Intvect(1,0,0);
            case 2 : return Intvect(0,1,0);
            case 3 : return Intvect(0,0,1);
            case 4 : return Intvect(0,1,1);
            case 5 : return Intvect(1,0,1);
            case 6 : return Intvect(1,1,0);
            case 7 : return Intvect(1,1,1);
            default: t.FatalError( "Illegal data centering code: %d.", dc );
        }
    }
    return Intvect(-1,-1,-1); // never gets here
}

int DataCenteringIntvect2Int( Intvect const & iv, int spaceDim )
{
    Trace t("DataCenteringIntvect2Int()");
    if( iv == Intvect(0,0,0) ) return 0;
    if( iv == Intvect(1,0,0) ) return 1;
    if( iv == Intvect(0,1,0) ) return 2;
    if( iv == Intvect(0,0,1) ) return 3;
    if( iv == Intvect(0,1,1) ) return 4;
    if( iv == Intvect(1,0,1) ) return 5;
    if( iv == Intvect(1,1,0) )
    {
        if( spaceDim == 2 ) return 7;
        else                return 6;
    }
    if( iv == Intvect(1,1,1) ) return 7;
    /* else */
    t.FatalError( "Illegal datacentering: (%d,%d,%d).", iv.i(), iv.j(), iv.k());
    return -1; // never gets here.
}


/** Arg centering is one of our hdf5 data-centering codes.  See the comments
 *  above for what the codes mean.
*/
bool
IsLegalDataCentering( int centering )
{
    Trace t("ChomboHDF5::IsLegalDataCentering()");
    return (centering>=0 && centering<=7);
}


//
// Disable/enable HDF5's default printing of errors.  Useful for when you want
// to call H5Gget_objinfo() to find out if something exists.
//
H5E_auto_t g_defaultHDF5ErrorHandler;
int * g_defaultHDF5ErrorHandlerClientData(0);
void DisableHDF5ErrorPrinting()
{
    // First, save the defaults so you can later reenable them -- if you haven't
    // already saved the defaults.
    if( ! g_defaultHDF5ErrorHandlerClientData )
    {
        H5Eget_auto( &g_defaultHDF5ErrorHandler,
                    (void * *)&g_defaultHDF5ErrorHandlerClientData );
    }
    H5Eset_auto( 0,0 );
}

void ReEnableHDF5ErrorPrinting()
{
    if( g_defaultHDF5ErrorHandlerClientData )
    { // Otherwise, it hasn't been disabled.
        H5Eset_auto( g_defaultHDF5ErrorHandler,
                     g_defaultHDF5ErrorHandlerClientData );
    }
}
