#
#   _______              __
#  / ___/ /  ___  __ _  / /  ___
# / /__/ _ \/ _ \/  ' \/ _ \/ _ \
# \___/_//_/\___/_/_/_/_.__/\___/ 
#
# This software is copyright (C) by the Lawrence Berkeley
# National Laboratory.  Permission is granted to reproduce
# this software for non-commercial purposes provided that
# this notice is left intact.
# 
# It is acknowledged that the U.S. Government has rights to
# this software under Contract DE-AC03-765F00098 between
# the U.S. Department of Energy and the University of
# California.
#
# This software is provided as a professional and academic
# contribution for joint exchange.  Thus it is experimental,
# is provided ``as is'', with no warranties of any kind
# whatsoever, no support, no promise of updates, or printed
# documentation.  By using this software, you acknowledge
# that the Lawrence Berkeley National Laboratory and
# Regents of the University of California shall have no
# liability with respect to the infringement of other
# copyrights by any part of this software.
#

# File: control_cameras.py
# Author: TDSternberg
# Created: 10/29/01

import types
import Tkinter
import Pmw

import anag_utils
import anag_megawidgets
from self_control import SelfControl

class ControlCameras( anag_megawidgets.SelfDescribingDialog, SelfControl ):
    """
    GUI control over clipping plane and camera parameters.
    """

    def getShortDescription(self):
        return "Cameras"
    def getLongDescription(self):
        return "Various camera parameters"


    def __init__( self, dep_dict, master=None, **kw ):
        anag_utils.funcTrace()
        anag_megawidgets.SelfDescribingDialog.__init__( self, master, **kw )
        self.configure( title='Camera parameters' )
        SelfControl.__init__( self, dep_dict,
          [
            {'name':'clipping_min'},  # EntryScale
            {'name':'clipping_max'},  # EntryScale
            {'name':'params_display', 'initval':None}
          ])
        self.decls = anag_utils.Declarations( "decls", instance=self )
        self.configure( title = "Cameras" )
        self.dialog_inventory.registerDialog( self )

    def _initForFirstHDF5( self ):
        anag_utils.funcTrace()

        Tkinter.Label( self.interior(), text="Clipping Range" ).pack()

        range_min = round(
            self.vtk_cameras.getClippingRangeMin()*0.9 - 0.5 )
        range_max = round(
            self.vtk_cameras.getClippingRangeMax()*1.1 + 0.5 )

        
        # As long as we're not offering a link to these widgets from the menu
        # system, let's not initialize them.  Initializing them seems to have an
        # effect on the actual clipping.
        """
        self.clipping_min = anag_megawidgets.EntryScale(
            self.interior(),
            from_ = range_min,
            to = range_max,
            length = 100,
            scale_normal_resolution = 0.1,
            scale_callback = self.setClippingRangeMin,
            button_text = 'min',
            relief = Tkinter.GROOVE, borderwidth=2,
            dep_dict = {'saved_states':self.saved_states }
            )
        self.clipping_max = anag_megawidgets.EntryScale(
            self.interior(),
            from_ = range_min,
            to = range_max,
            length = 100,
            scale_normal_resolution = 0.1,
            scale_callback = self.setClippingRangeMax,
            button_text = 'max',
            relief = Tkinter.GROOVE, borderwidth=2,
            dep_dict = {'saved_states':self.saved_states }
            )
        clipping_range = ( self.vtk_cameras.getClippingRangeMin(),
                           self.vtk_cameras.getClippingRangeMax() )
        self.clipping_min.setScaleValue( clipping_range[0] )
        self.clipping_max.setScaleValue( clipping_range[1] )

        self.clipping_min.pack()
        self.clipping_max.pack()
        """


    def _refresh( self ):
        """ Overrides SelfControl._refresh().  See comments there. """
        anag_utils.funcTrace()

        self.setClippingRangeMin(
            self.vtk_cameras.getCamera().GetClippingRange()[0] )
        self.setClippingRangeMax(
            self.vtk_cameras.getCamera().GetClippingRange()[1] )


    def cleanup( self ):
        anag_utils.funcTrace()
        if self.params_display:
            self.params_display.withdraw()
            self.params_display.cleanup()
        self.dialog_inventory.unregisterDialog( self )
        self.withdraw()


    def setClippingRangeMin( self, x ):
        #self.clipping_max.configure( scale_soft_min=x )
        self.vtk_cameras.setClippingRange( range_min=x )


    def setClippingRangeMax( self, x ):
        #self.clipping_min.configure( scale_soft_max=x )
        self.vtk_cameras.setClippingRange( range_max=x )


    def showParams( self ):
        if not self.params_display:
            self.params_display = ParamsDisplay( dep_dict = {
              'vol_tk_interactor':self.vol_tk_interactor,
              'vtk_cameras':self.vtk_cameras,
              'saved_states':self.saved_states })
            self.dialog_inventory.registerDialog( self.params_display )
        self.params_display.show()


class ParamsDisplay( SelfControl, Pmw.Dialog ):
    """
    Display the camera position, focal point, etc -- all the things a user would
    need to know if he wanted to do vol_tk_interactor stuff programatically.
    """
    def __init__( self, dep_dict, master=None, **kw ):
        anag_utils.funcTrace()
        SelfControl.__init__( self, dep_dict, metadata=
            [ {'name':'camera_position'},
              {'name':'axes_camera_position'},
              {'name':'focal_point'},
              {'name':'world_point'},
              {'name':'clipping_range'},
              {'name':'display_point'},
              {'name':'parallel_scale'},
              {'name':'view_up'},
              {'name':'axes_view_up'} ])
        Pmw.Dialog.__init__( self, master, **kw )
        self.initialiseoptions(Pmw.Dialog)
        self.configure( buttons=('close', 'dump'), title='Camera parameters',
                        command=self._buttonHandler )

        #self.withdraw()

        self.decls = anag_utils.Declarations( 'decls', instance=self )
        decls = anag_utils.Declarations( 'decls', 'self', 'master', 'kw',
            'dep_dict', 'multi_entry' )

        # The data gets displayed in noneditable EntryFields.
        self.camera_position = MultiEntry( n_entries=3,
            master=self.interior(), label_text='camera position' )
        self.axes_camera_position = MultiEntry( n_entries=3,
            master=self.interior(), label_text='axes camera position' )
        self.focal_point = MultiEntry( n_entries=3,
            master=self.interior(), label_text='focal point' )
        self.world_point = MultiEntry( n_entries=4,
            master=self.interior(), label_text='world point' )
        self.display_point = MultiEntry( n_entries=3,
            master=self.interior(), label_text='display point' )
        self.view_up = MultiEntry( n_entries=3,
            master=self.interior(), label_text='view up' )
        self.axes_view_up = MultiEntry( n_entries=3,
            master=self.interior(), label_text='axes view up' )
        self.parallel_scale = MultiEntry( n_entries=1,
            master=self.interior(), label_text='parallel scale' )
        self.clipping_range = MultiEntry( n_entries=2,
            master=self.interior(), label_text='clipping range', is_writable=1)
        self.clipping_range.setCommand( self.vtk_cameras.setClippingRange )
        for multi_entry in (self.camera_position, self.axes_camera_position,
            self.focal_point, self.world_point, self.display_point,
            self.view_up, self.axes_view_up,
            self.clipping_range, self.parallel_scale ):
            multi_entry.pack( anchor='w')

        self._refresh()
        decls.memberFunctionAudit( self )


    def _initForFirstHDF5( self ):
        anag_utils.funcTrace()


    def _refresh( self ):
        self.registerCallback(
            self.vol_tk_interactor.getNotifierCameraPosition(),
            lambda multi, extra_info, entries=self.camera_position:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierAxesCameraPosition(),
            lambda multi, extra_info, entries=self.axes_camera_position:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierFocalPoint(),
            lambda multi, extra_info, entries=self.focal_point:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierWorldPoint(),
            lambda multi, extra_info, entries=self.world_point:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierDisplayPoint(),
            lambda multi, extra_info, entries=self.display_point:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierViewUp(),
            lambda multi, extra_info, entries=self.view_up:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierAxesViewUp(),
            lambda multi, extra_info, entries=self.axes_view_up:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierParallelScale(),
            lambda multi, extra_info, entries=self.parallel_scale:
                entries.set( multi ))
        self.registerCallback(
            self.vol_tk_interactor.getNotifierClippingRange(),
            lambda multi, extra_info, entries=self.clipping_range:
                entries.set( multi ))

        self.camera_position.set( self.vol_tk_interactor.getCameraPosition() )
        self.focal_point.set( self.vol_tk_interactor.getFocalPoint() )
        self.world_point.set( self.vol_tk_interactor.getWorldPoint() )
        self.display_point.set( self.vol_tk_interactor.getDisplayPoint() )
        self.view_up.set( self.vol_tk_interactor.getViewUp() )
        self.parallel_scale.set( self.vol_tk_interactor.getParallelScale() )
        self.clipping_range.set( self.vol_tk_interactor.getClippingRange() )


    def cleanup( self ):
        anag_utils.funcTrace()
        self.unregisterCallbacks()
        self.zeroSelfVariables()


    def _buttonHandler( self, button_name ):
        if button_name == 'dump':
            self.vol_tk_interactor.printParams()
        else:  # Covers 'close' button, as well as window close widget.
            self.withdraw()


class MultiEntry( Tkinter.Frame ):
    """
    A number of Entry widgets -- typically representing x and y, and maybe z --
    plus a label.
    Methods to set the widgets one at a time or (from a tuple) all together.
    """
    def __init__( self, master, n_entries, label_text, is_writable=None,
                  command=None, **kw ):
        anag_utils.funcTrace()
        Tkinter.Frame.__init__( self, master, **kw )
        
        self.is_writable = is_writable

        Tkinter.Label( self, text=label_text, width=20 ).pack( side='left' )
        self.entry = []
        for i in range(0,n_entries):
            self.entry.append(Pmw.EntryField(self))
            self.entry[i].component('entry').configure( state='disabled' )
            self.entry[i].pack( side='left' )


    def setCommand( self, function ):
        """
        Every time there's a carriage return in any of the elements of
        self.entry, apply the contents of all the elements of self.entry as
        separate parameters to arg function.
        """
        anag_utils.funcTrace()
        for e in self.entry:
            e.configure( command = lambda f=function, self=self:
                self._applyFunctionToAllEntries( f ))


    def _applyFunctionToAllEntries( self, function ):
        args = []
        for item in self.entry:
            args.append( float(item.get()) )
        apply( function, list(args) )


    def set( self, text_multi ):
        anag_utils.funcTrace()
        if type(text_multi) == types.TupleType:
            for i in range(0,len(self.entry)):
                self.entry[i].component('entry').configure( state='normal' )
                self.entry[i].setentry( str(text_multi[i]) )
                if not self.is_writable:
                    self.entry[i].component('entry'
                                           ).configure( state='disabled' )
        else: # special case for parallel_scale
            self.entry[0].component('entry').configure( state='normal' )
            self.entry[0].setentry( str(text_multi) )
            self.entry[0].component('entry').configure( state='disabled' )
