#
#   _______              __
#  / ___/ /  ___  __ _  / /  ___
# / /__/ _ \/ _ \/  ' \/ _ \/ _ \
# \___/_//_/\___/_/_/_/_.__/\___/ 
#
# This software is copyright (C) by the Lawrence Berkeley
# National Laboratory.  Permission is granted to reproduce
# this software for non-commercial purposes provided that
# this notice is left intact.
# 
# It is acknowledged that the U.S. Government has rights to
# this software under Contract DE-AC03-765F00098 between
# the U.S. Department of Energy and the University of
# California.
#
# This software is provided as a professional and academic
# contribution for joint exchange.  Thus it is experimental,
# is provided ``as is'', with no warranties of any kind
# whatsoever, no support, no promise of updates, or printed
# documentation.  By using this software, you acknowledge
# that the Lawrence Berkeley National Laboratory and
# Regents of the University of California shall have no
# liability with respect to the infringement of other
# copyrights by any part of this software.
#

# File: vtk_cameras.py
# Author: TDSternberg
# Created: 8/23/01

import anag_utils
from self_control import SelfControl

class VtkCameras( SelfControl ):
    def __init__(self, dep_dict):
        SelfControl.__init__( self, dep_dict,
            [
                {'name':'camera', 'get':1 },
                {'name':'light', 'get':1 },
                {'name':'axes_camera', 'get':1 },
                {'name':'axes_light', 'get':1 },
                {'name':'clipping_range_min', 'get':1},
                {'name':'clipping_range_max', 'get':1}
            ] )
        self.decls = anag_utils.Declarations( "decls", instance=self  )


    def _initForFirstHDF5( self ):
        anag_utils.funcTrace()
        decls = anag_utils.Declarations( "self", "decls", "is_new_hdf5",
            "lights", "range_", "minrange", "maxrange", "mid"
            )

        if self.cmd_line.getNoVtk():
            return

        self.vtk_vtk.getRenWin().Render() # Gotta call this, or else light==None

        self.camera = self.vtk_vtk.getRenderer().GetActiveCamera()

        lights = self.vtk_vtk.getRenderer().GetLights()
        lights.InitTraversal()
        self.light = lights.GetNextItem()
        self.camera.SetPosition( 0, 0, 1000 )
        self.camera.SetViewUp( 0, 1, 0 )
        self.camera.SetFocalPoint( 0, 0, 0 )
        self.camera.ComputeViewPlaneNormal()
        self.vtk_vtk.getRenderer().ResetCamera()

        if( self.vtk_data.is2DMode() ):
            self.camera.ParallelProjectionOn()
            # At this point, the clipping range has been set automatically by
            # vtk.  But from here on we take control of it.
            range_ = self.camera.GetClippingRange()
            minrange = range_[0]
            maxrange = range_[1]
            mid = ( minrange * maxrange ) ** 0.5
            self.setClippingRange( minrange / 4.0, maxrange * 4.0 )
            self.camera.SetParallelScale( 0.18 * mid )
        else:
            self.camera.Elevation( 20.0 )
            self.camera.Azimuth( -10.0 )
            self.camera.Dolly( 1.2 )
        
            # At this point, the clipping range has been set automatically by
            # vtk.  But from here on we take control of it.
            range_ = self.camera.GetClippingRange()
            self.setClippingRange( range_[0] / 4.0, range_[1] * 4.0 )

        # Deal with the little axes display.
        if( (not self.vtk_data.is2DMode())
        and (self.cmd_line.getUseRenderWidget() == 1)):
            self._initAxes()

        decls.memberFunctionAudit(self)


    def _refresh( self ):
        anag_utils.funcTrace()
        if self.cmd_line.getNoVtk():
            return


    def cleanup( self ):
        anag_utils.funcTrace()


    def _initAxes( self ):
        anag_utils.funcTrace()
        decls = anag_utils.Declarations( 'self', 'decls', 'lights', 'range_',
            'camera_focus')

        self.vtk_vtk.getAxesRenWin().Render()

        self.axes_camera = self.vtk_vtk.getAxesRenderer().GetActiveCamera()
        lights = self.vtk_vtk.getAxesRenderer().GetLights()
        lights.InitTraversal()

        self.axes_light = lights.GetNextItem()

        camera_focus = self.axes_camera.GetFocalPoint()
        self.axes_camera.SetFocalPoint( camera_focus )
        self.axes_camera.SetPosition( camera_focus[0], camera_focus[1], 1000 )
        self.axes_camera.SetViewUp( 0, 1, 0 )
        self.axes_camera.ComputeViewPlaneNormal()
        self.vtk_vtk.getRenderer().ResetCamera()
        
        self.axes_camera.Elevation( 20.0 )
        self.axes_camera.Azimuth( -10.0 )
        self.axes_camera.Dolly( 109 )
        
        self.axes_light.SetPosition( self.axes_camera.GetPosition() )
        self.axes_light.SetFocalPoint( self.axes_camera.GetFocalPoint() )

        range_ = self.axes_camera.GetClippingRange()
        self.axes_camera.SetClippingRange( 0.1, 20 )

        decls.memberFunctionAudit(self)


    def setLightIntensity( self, x ):
        assert(0<=x<=1)
        self.light.SetIntensity(x)


    def setClippingRange( self, range_min=None, range_max=None ):
        anag_utils.funcTrace()
        if   range_min == None  and  range_max != None:
            self._setClippingRangeMax( range_max )
        elif range_min != None  and  range_max == None:
            self._setClippingRangeMin( range_min )
        elif range_min != None  and  range_max != None:
            self._setClippingRangeMin( range_min )
            self._setClippingRangeMax( range_max )
        else:
            anag_utils.Fatal( 'At least one of range_min and range_max must be '
                              'other than None.' )

        self.vtk_vtk.render()

    def _setClippingRangeMin( self, x ):
        if not self.clipping_range_max: # Not initialized yet
            (dummy,self.clipping_range_max) = self.camera.GetClippingRange()
        self.camera.SetClippingRange( x/4.0, self.clipping_range_max*4.0 )
        self.clipping_range_min = x
        self.vtk_vtk.render()

    def _setClippingRangeMax( self, x ):
        if not self.clipping_range_min: # Not initialized yet
            (self.clipping_range_min,dummy) = self.camera.GetClippingRange()
        self.camera.SetClippingRange( self.clipping_range_min/4.0, x*4 )
        self.clipping_range_max = x
        self.vtk_vtk.render()
