#include "cmdline_base.h"
#include "Trace.h"
#include <fstream>
#include <iostream>
#include <cstdio> 
#include <cstdlib> 
#include <cassert>
#include <map>
#include <string>
#include <vector>
#include <sys/types.h> 
#include <sys/stat.h>
#include <time.h>

using std::vector;
using std::string; 
using std::map;
using std::cout;
using std::cerr;
using std::endl;

//----------------------------------------------------------------- 
// 
// class Cmdline 
// 
// Process configuration data from the command line and/or cmdline file.
//  
//----------------------------------------------------------------- 

/** Initialize the Cmdline object from the master configuration file.
 *  Then let any command-line override what you got from the file.
 *  An STL map acts as the one-stop reference and interface to all
 *  the parameters.
*/
CmdlineBase::CmdlineBase()
{ 
    m_argmap["debug_level"] = new SetInt(&s_debug_level, "");
}


/** Supports feature whereby configuration is sensitive to run-time
 *   modifications to cmdline file.
*/
void
CmdlineBase::SaveCmdlineFileTimestamp()
{
    if( !m_cmdlineFileExists )
    {
        return;
    }

    // Save timestamp of cmdline file.
    struct stat mystat;
    stat( m_cmdlineFilename.c_str(), &mystat );
    m_lastTouch = mystat.st_mtime;
}


/** Cmdline file is assumed to reside in $HOME directory.
 *  New: It's not obligatory to have a cmdline file, and even if you have one
 *       it doesn't have to mention all the cmdline variables.
 */
void
CmdlineBase::CheckForCmdlineFile( string filename )
{
    Trace t("CmdlineBase::CheckForCmdlineFile()");
    m_cmdlineFilename = filename;

    // See if it exists.
/*
    //FIXME: This stat stuff fails under efence, on c532352-a (gcc2.95.3)
    //it doesn't fail on hepburn, either against gcc2.95.3 or gcc3.2.2.  It also
    //doesn't fail if I set the env variable EF_PROTECT_BELOW=1.  It fails
    //in the ChomboHDF5-unittest but also the misctest executable that does a
    //minimal amount of stuff -- just as long as it uses VisualizableDataset
    //templates.  This, as of September 30, 2003.
    struct stat mystat;
    stat( m_cmdlineFilename.c_str(), &mystat );
    if( ! S_ISREG(mystat.st_mode) )
*/
    std::ifstream infile( m_cmdlineFilename.c_str() );
    if( !infile )
    {
        m_cmdlineFileExists = false;
    } else
    {
        m_cmdlineFileExists = true;
    }
}


/** Check if m_cmdlineFilename has been touched.  If yes, then reset cmdline
 *  parameters.  Note this will override anything originally set from the
 *  command line (and this is perhaps counterintuitive, because in
 *  Cmdline::Cmdline(), the command line settings override the cmdline file
 *  settings.  So it's probably a good idea to set your params from the cmdline
 *  file, rather than from the command line.
 *
 *  If there's no cmdline file (precisely: if there was no cmdline file at
 *  start-up), then this is a no-op.
*/
void
CmdlineBase::Reload()
{
    if( ! m_cmdlineFileExists )
    {
        return;
    }
    struct stat mystat;
    stat( m_cmdlineFilename.c_str(), &mystat );
    if( mystat.st_mtime > m_lastTouch )
    {
        LoadDefaultsFromFile();
        m_lastTouch = mystat.st_mtime;
    }
}

/** Cmdlineuration file must conform to the following format:
 *  Comments begin with #
 *  Blank lines are ok.
 *  All other lines should indicate the value of a parameter, in name=value
 *    format.  The name should start in the first column of the line, and
 *    there should be no blanks around the = sign.
 *  For string types, it's ok to put nothing after the = sign; this indicates
 *    a default value of "".
 *  
 *  The cmdline file is optional.  When it does exist, it does not need to
 *  mention every variable.
*/
void CmdlineBase::LoadDefaultsFromFile()
{
    Trace t("CmdlineBase::LoadDefaultsFromFile()");
    if( !m_cmdlineFileExists )
    {
        return;
    }

    // Open cmdline file
    std::ifstream infile( m_cmdlineFilename.c_str() );
    if( !infile )
    {
        t.FatalError( "Shouldn't have gotten here." );
    }

    // Read cmdline file and set data members.
    char linebuf[82];
    vector<string> all_param_names;
    while( !infile.eof() )
    {
        infile.getline( linebuf,81);
        if( strlen( linebuf ) > 80 )
        {
            std::cerr << m_cmdlineFilename 
                      << " lines must not exceed 80 columns.  Found one at "
                      << strlen(linebuf) << ", and here it is:\n"
                      << "|" << linebuf << "|" << "\n";
            exit(1);
        }

        // If the line starts with a character other than #, then interpret it
        // and set the corresponding data member.
        if( !isalpha(linebuf[0]) ) ;
        else
        {
            if( m_argmap.find(ArgName(linebuf)) != m_argmap.end() )
            {
                m_argmap[ArgName(linebuf)]->set(ArgVal(linebuf));
                if( ArgName(linebuf) == "debug_level" )
                {
                    DebugLevel( atoi(ArgVal(linebuf).c_str()) );
                }
                all_param_names.push_back( ArgName(linebuf) );
            }
        }
    }
}        
    

/** Command-line options all take the form of name=value, e.g.
 *  "lambda=0.3", "logfilelist=logs.dat".
 *
 *  The only other valid thing for the command line is "-h" or "--help".
*/
void CmdlineBase::ParseCommandLine( int argc, char *argv[] )
{
    // Print help, if argv is -h or --help.
    if( argc==2 && ( string(argv[1])=="-h" || string(argv[1])=="--help" ))
    {
        for( map<string,SetType*>::iterator i=m_argmap.begin();
                                            i!=m_argmap.end();
                                            i++ )
        {
            std::cerr << i->first << "       # " << i->second->getDescription() 
                 << std::endl;
        }
        cout << "**********************************************\n";
        exit(0);
    }

    // Go through all of argv[].
    for( int i=1;i<argc;i++ )
    {
        if( m_argmap.find( ArgName(argv[i]) ) == m_argmap.end() )
        {
            std::cerr << "***Error: " << ArgName(argv[i]) << " is not a legal "
                 << "command-line argument.  " << std::endl
                 << "Use -h option for help." << std::endl;
            exit(1);
        }
        m_argmap[ArgName(argv[i])]->set(ArgVal(argv[i]));
        std::cerr << "cmd-line-option " << ArgName(argv[i]) << "=" 
             << ArgVal(argv[i]) << std::endl;
        if( ArgName(argv[i]) == "debug_level" )
        {
            DebugLevel( atoi(ArgVal(argv[i]).c_str()) );
        }
    } 
} 
 
/** Arguments -- both at the command line and in the cmdline file --
 *  have the name=value format.
 *  This function returns the first part -- the name.
*/
string CmdlineBase::ArgName( string str )
{
    char* buf = new char[ str.length()+1 ];
    strcpy( buf, str.c_str() );
    string result( strtok( buf, "=" ) );
    delete[] buf;
    return result;
}

/** Arguments -- both at the command line and in the cmdline file --
 *  have the name=value format (no spaces allowed around the '=' sign!)
 *  This function returns the second part -- the value.  If it's a string, put
 *  double quotes around it; we'll strip them off.
*/
string CmdlineBase::ArgVal( string str )
{
    string result;
    char * buf = new char[ str.length()+1 ];
    strcpy( buf, str.c_str() );
    char * tok = strtok( buf, "=" );  assert( tok );
    tok = strtok( NULL, " \t\0\n#" ); assert( tok );

    // Strip off surrounding '"', if any.
    if( (tok[0]=='"') && (tok[strlen(tok)-1]=='"') )
    {
        tok[strlen(tok)-1] = 0;
        result = string( tok+1 );
    } else
    {
        result = string( tok );
    }

    delete[] buf;
    return result;
}


/*static*/ void
CmdlineBase::DebugLevel( int value )
{
    Trace t("CmdlineBase::DebugLevel()");
    t.Info("");
    s_debug_level = value;
} 

/*static*/int
CmdlineBase::DebugLevel()
{
    return s_debug_level;
}

int CmdlineBase::s_debug_level = 3;
